/**
 * \file GkrLeg.cpp
 * \brief Drive one leg of the Geckrobot thanks to the HerkuleXLib
 * \author Benoit Puel (contact@benoitpuel.com)
 * \version 1.0
 * \date 26 December 2015
 * \copyright This project is released under the GNU GPL V3.
 */

#include "GkrLeg.h"

GkrLeg::GkrLeg(HkxGroupPosControl& servos, HkxPrint& print, int16_t shoulderPosition[], int16_t legDistances[]) : _servos(servos), _print(print), _className("GkrLeg::"){
  if(servos.getNbServos()!=3){
    errorPrint(F("GkrLeg > input group of servos shall contain 3 servos"));
    return;
  }
  memcpy(_shoulderPosition, shoulderPosition, sizeof(_shoulderPosition)); 
  memcpy(_legDistances, legDistances, sizeof(_legDistances)); 
  _pbDetected = false;

  updateCurrentPosition();
}

uint8_t GkrLeg::updateCurrentPosition(){
  for(int i=0 ; i<3 ; i++){
    uint8_t error = _servos.getServo(i).getBehaviour(HKX_NO_VALUE, HKX_NO_VALUE, &_currentJointPosition[i], HKX_NO_VALUE, HKX_NO_VALUE, HKX_NO_VALUE, HKX_NO_VALUE, HKX_NO_VALUE);
    if(error>0){
      return error;
    }
  }
  jointToSpaceConversion(_currentJointPosition, _currentSpacePosition);
  return 0;
}

bool GkrLeg::spaceToJointConversion(const int16_t spacePosition[], int16_t jointPosition[]){
  // pre-processing
  if (_legDistances[1] == 0 || _legDistances[2] == 0){
    return false;
  }
  float A = sqrt(sq((float)spacePosition[0] - _shoulderPosition[0]) + sq((float)spacePosition[1] - _shoulderPosition[1])); // in 10^-1 mm
  float cosTheta3 = (sq((float)A - _legDistances[0]) + sq((float)spacePosition[2] - _shoulderPosition[2]) - sq((float)_legDistances[1]) - sq((float)_legDistances[2])) / (2 * (float)_legDistances[1] * _legDistances[2]);
  if(abs(cosTheta3) > 1){
    //errorPrint((String)"spaceToJointConversion > cosTheta3 = "+cosTheta3);
    return false;
  }
  
  // jointPosition[0] / theta1
  jointPosition[0] = radToDeg(atan2(_shoulderPosition[0] - spacePosition[0], spacePosition[1] - _shoulderPosition[1])); // in 10^-1 degree
  
  // jointPosition[2] / theta3
  float theta3 = - acos(cosTheta3); // in rad
  float sinTheta3 = sin(theta3);
  jointPosition[2] = radToDeg(theta3); // in 10^-1 degree
  
  // jointPosition[1] / theta2
  if(abs(sinTheta3) < 0.001){ // sinTheta3 = 0 (to avoid division by zero)
      jointPosition[1] = radToDeg(atan2((float)spacePosition[2] - _shoulderPosition[2], A - _legDistances[0])); // in 10^-1 degree
  } else {
      float tmp1 = (float)_legDistances[1] + _legDistances[2]*cosTheta3; // in 10^-1 mm
      float tmp2 = (float)_legDistances[2]*sinTheta3; // in 10^-1 mm
      float cosTheta2 = (tmp1 * (A - _legDistances[0]) + tmp2 * (spacePosition[2] - _shoulderPosition[2])) / (sq(tmp1) + sq(tmp2));
      float sinTheta2 = (tmp1 * cosTheta2 - A + _legDistances[0]) / tmp2;
      jointPosition[1] = radToDeg(atan2(sinTheta2, cosTheta2)); // in 10^-1 degree
  }

  return true;
}

void GkrLeg::jointToSpaceConversion(const int16_t jointPosition[], int16_t spacePosition[]){
  // pre-processing
  float cosTheta1 = cos(degToRad(jointPosition[0]));
  float sinTheta1 = sin(degToRad(jointPosition[0]));
  float cosTheta2 = cos(degToRad(jointPosition[1]));
  float sinTheta2 = sin(degToRad(jointPosition[1]));
  float cosTheta23 = cos(degToRad(jointPosition[1]+jointPosition[2]));
  float sinTheta23 = sin(degToRad(jointPosition[1]+jointPosition[2]));
  
  
  float temp1 = sinTheta2*_legDistances[2];
  float temp2 = _legDistances[1] - cosTheta2*_legDistances[2];
  
  // deltaY and deltaZ
  float deltaY = _legDistances[0] + _legDistances[1] * cosTheta2 + _legDistances[2] * cosTheta23;
  float deltaZ = _legDistances[1] * sinTheta2 + _legDistances[2] * sinTheta23;

  // processing
  spacePosition[0] = _shoulderPosition[0] - deltaY * sinTheta1; // X
  spacePosition[1] = _shoulderPosition[1] + deltaY * cosTheta1; // Y
  spacePosition[2] = _shoulderPosition[2] + deltaZ; // Z
}

// maxTorque en N.mm
// forces en N
void GkrLeg::getMaxTorque(const int16_t jointPosition[], const int16_t forces[], int16_t maxTorque[]){
  // pre-processing
  float cosTheta1 = cos(jointPosition[0]);
  float sinTheta1 = sin(jointPosition[0]);
  float cosTheta2 = cos(jointPosition[1]);
  float sinTheta2 = sin(jointPosition[1]);
  float cosTheta23 = cos(jointPosition[1] + jointPosition[2]);
  float sinTheta23 = sin(jointPosition[1] + jointPosition[2]);
  float tmp1 = forces[0]*sinTheta1 - forces[1]*cosTheta1;
  
  // processing
  maxTorque[0] = - (_legDistances[0] + _legDistances[1]*cosTheta2 + _legDistances[2]*cosTheta23)*(forces[0]*cosTheta1 + forces[1]*sinTheta1)/10.0;
  maxTorque[2] = _legDistances[2] * (tmp1*sinTheta23 + forces[2]*cosTheta23)/10.0;
  maxTorque[1] = (_legDistances[1] * (tmp1*sinTheta2 + forces[2]*cosTheta2) + maxTorque[2])/10.0;
  //infoPrint((String)"getMaxTorque > torque0="+maxTorque[0]+" torque1="+maxTorque[1]+" torque2="+maxTorque[2]);
}

// torque in N.mm, velocity in 10^-1deg/s
uint8_t GkrLeg::applyMaxTorque(const int16_t maxTorque[], const int16_t velocity[]){
  String message = "applyMaxTorque >";
  uint16_t inputVoltage;
  int16_t maxPWM;
  const uint16_t motorConstantTorque = 2818; // Value depend on PID controle settings
  const uint16_t motorConstantVelocity = 1275; // Value depend on PID controle settings
  for (uint8_t i = 0 ; i < 3 ; i++){
    uint8_t error = _servos.getServo(i).getBehaviour(&inputVoltage, HKX_NO_VALUE, HKX_NO_VALUE, HKX_NO_VALUE, HKX_NO_VALUE, HKX_NO_VALUE, HKX_NO_VALUE, HKX_NO_VALUE);
    maxPWM = maxPWM==0?HKX_MAX_PWM:max(abs(((motorConstantVelocity*velocity[i]*3142)/180 + motorConstantTorque*maxTorque[i])/(1023 * inputVoltage * 1000)),100);
    _servos.getServo(i).setLoad(HKX_NO_VALUE, HKX_NO_VALUE, HKX_NO_VALUE, HKX_NO_VALUE, HKX_NO_VALUE, maxPWM);
    message += (String)" servo"+i+" PWM = "+maxPWM;
  }
  infoPrint(message);
}

uint8_t GkrLeg::moveSpace(int16_t spacePosition[], uint32_t playTime, HkxMaybe<hkxLEDControl> tabLedNoValue[], int16_t forces[]){
  int16_t jointPosition[3];
  if(!spaceToJointConversion(spacePosition, jointPosition)){// check if out the boundaries of the leg
    errorPrint(F("moveSpace > The end position is out of the leg limits"));
    return 10;
  }

  String message;
  int16_t initialPosition[3], stepPosition[3], maxTorque[3], velocity[3], previousJointPosition[3];
  updateCurrentPosition();
  memcpy(initialPosition, _currentSpacePosition, sizeof(_currentSpacePosition));
  
  // Nb of step evaluated for the distance
  int32_t nbStepDist = ceil(sqrt(sq((int32_t)initialPosition[0]-spacePosition[0]) + sq((int32_t)initialPosition[1]-spacePosition[1]) + sq((int32_t)initialPosition[2]-spacePosition[2]))/20.0); // mesh with 2 mm step
  // Nb of step evaluated for the time
  int16_t maxStepTime = playTime/50; // mesh to get step of 50 ms
  int16_t minStepTime = 1+playTime/1400; // mesh to ensure to generate trajectories < 2845 ms
  int16_t nbStep = max(min(nbStepDist,maxStepTime),minStepTime); // use the smallest nb of step, but at least minStepTime !
  infoPrint((String)"moveSpace > nbStep: "+nbStep);

  // Preprocessing 
  int32_t xDist = spacePosition[0] - initialPosition[0];
  int32_t yDist = spacePosition[1] - initialPosition[1];
  int32_t zDist = spacePosition[2] - initialPosition[2];
  uint32_t startTimer = millis();
  uint32_t stepTime = startTimer, nextStepTime = startTimer, nextNextStepTime = startTimer+playTime/nbStep;

  //uint32_t stepTime = startTimer, nextStepTime = startTimer+playTime/nbStep;
  for(int i = 0; i < nbStep ; i++){
    stepPosition[0] = initialPosition[0]+((int32_t)xDist*min(i+2,nbStep))/nbStep;
    stepPosition[1] = initialPosition[1]+((int32_t)yDist*min(i+2,nbStep))/nbStep;
    stepPosition[2] = initialPosition[2]+((int32_t)zDist*min(i+2,nbStep))/nbStep;
    //errorPrint((String)"moveSpace > stepX = " + stepPosition[0] + " stepY = " + stepPosition[1] + " stepZ = " + stepPosition[2]);
    
    memcpy(previousJointPosition, jointPosition, sizeof(jointPosition));
    if(!spaceToJointConversion(stepPosition, jointPosition)){
      errorPrint("The step position cannot be reached");
    }
    //errorPrint((String)"moveSpace > stepTh0 = " + jointPosition[0] + " stepTh1 = " + jointPosition[1] + " stepTh2 = " + jointPosition[2]);
    
    stepTime = nextStepTime;
    nextStepTime = nextNextStepTime;
    nextNextStepTime = min(startTimer+(playTime*(i+2))/nbStep, startTimer+playTime);
    //nextStepTime = min(startTimer+(playTime*(i+1))/nbStep, startTimer+playTime);
    //errorPrint((String)"moveSpace > stepTime = " + stepTime);
    
    /*getMaxTorque(_currentJointPosition, forces, maxTorque);
    for(uint8_t j = 0; j < 3 ; j++){
      velocity[j] = (jointPosition[j] - previousJointPosition[j])*1000/(nextNextStepTime-nextStepTime);
    }
    applyMaxTorque(maxTorque, velocity);*/
    while(millis() < stepTime){ // wait for next step
    }
    updateCurrentPosition();
    _servos.moveSyncAllPosition(jointPosition, nextNextStepTime-stepTime, tabLedNoValue, false);
    //_servos.moveSyncAllPosition(jointPosition, nextStepTime-stepTime, tabLedNoValue, false);
    
    //errorPrint((String)"moveSpace > posX = " + _currentSpacePosition[0] + " posY = " + _currentSpacePosition[1] + " posZ = " + _currentSpacePosition[2]);
    //errorPrint((String)"moveSpace > posTh0 = " + _currentJointPosition[0] + " posTh1 = " + _currentJointPosition[1] + " posTh2 = " + _currentJointPosition[2]);

  }
  while(!_servos.areInPosition() && millis() - startTimer < playTime + HKX_MOVE_OVERTIME){}; // wait servos are arrived or being overtime
  _servos.setAllTorqueLEDControl(HKX_TORQUE_FREE, HKX_LED_OFF);
  updateCurrentPosition();
}

uint8_t GkrLeg::moveJoints(int16_t jointPosition[], uint32_t playTime, HkxMaybe<hkxLEDControl> tabLedNoValue[]){
  int16_t initialPosition[3], stepPosition[3];
  int16_t nbStep = 1+playTime/1400; // mesh to ensure to generate trajectories < 2845 ms
  infoPrint((String)"moveSpace > nbStep: "+nbStep);
  updateCurrentPosition();
  memcpy(initialPosition, _currentJointPosition, sizeof(_currentJointPosition));
  
  int32_t xDist = jointPosition[0] - initialPosition[0];
  int32_t yDist = jointPosition[1] - initialPosition[1];
  int32_t zDist = jointPosition[2] - initialPosition[2];
  uint32_t startTimer = millis();
  uint32_t stepTime = startTimer, nextStepTime = startTimer, nextNextStepTime = startTimer+playTime/nbStep;
  
  for(int i = 0; i < nbStep ; i++){
    stepPosition[0] = initialPosition[0]+((int32_t)xDist*min(i+2,nbStep))/nbStep;
    stepPosition[1] = initialPosition[1]+((int32_t)yDist*min(i+2,nbStep))/nbStep;
    stepPosition[2] = initialPosition[2]+((int32_t)zDist*min(i+2,nbStep))/nbStep;
    
    stepTime = nextStepTime;
    nextStepTime = nextNextStepTime;
    nextNextStepTime = min(startTimer+(playTime*(i+2))/nbStep, startTimer+playTime);

    while(millis() < stepTime){ // wait for next step
    }
    updateCurrentPosition();
    _servos.moveSyncAllPosition(stepPosition, nextNextStepTime-stepTime, tabLedNoValue, false);
  }
  while(!_servos.areInPosition() && millis() - startTimer < playTime + HKX_MOVE_OVERTIME){}; // wait servos are arrived or being overtime
  _servos.setAllTorqueLEDControl(HKX_TORQUE_FREE, HKX_LED_OFF);
  updateCurrentPosition();
}

uint8_t GkrLeg::getCurrentSpacePosition(int16_t spacePosition[], bool update){
  if(update){
    updateCurrentPosition();
  }
  memcpy(spacePosition, _currentSpacePosition, sizeof(_currentSpacePosition)); 
  return 0;
}

uint8_t GkrLeg::getCurrentJointPosition(int16_t jointPosition[], bool update){
  if(update){
    updateCurrentPosition();
  }
  memcpy(jointPosition, _currentJointPosition, sizeof(_currentJointPosition)); 
  return 0;
}




